"""
Provides functional interface to git by running shell commands through the
subprocess module.
"""

import os
import locale
from subprocess import check_output, STDOUT, CalledProcessError
from pyhome.dircontext import dircontext

# Determine system encoding from locale
SYSENC = locale.getpreferredencoding()


class GitException(Exception):
    pass


def git(*args):
    """
    Run a git command.
    """

    # Construct command list
    cmd = ['git'] + list(args)

    # Attempt command and handle errors
    try:
        output = check_output(cmd, stderr=STDOUT)
    except OSError as e:
        raise GitException('git command not found')
    except CalledProcessError as e:
        raise GitException(e.output.decode(SYSENC).strip())
    finally:
        pass

    out = output.decode(SYSENC).strip()
    if len(out) > 0:
        print(out)


def reponame(url, name=None):
    """
    Determine a repo's cloned name from its URL.
    """
    if name is not None:
        return name
    name = os.path.basename(url)
    if name.endswith('.git'):
        name = name[:-4]
    return name


def init(parent, name):
    """
    Initialise a git repo.
    """

    subcmd = ['init', name]

    with dircontext(parent):
        git(*subcmd)


def clone(parent, url, name=None, submodules=True):
    """
    Clone a git repo.
    """

    subcmd = ['clone', url]

    if name is not None:
        subcmd.append(name)

    with dircontext(parent):
        git(*subcmd)

        if submodules:
            with dircontext(reponame(url, name)):
                git('submodule', 'update', '--init')


def pull(repo, submodules=True):
    """
    Update a repo.
    """
    with dircontext(repo):
        git('pull')
        if submodules:
            git('submodule', 'update', '--init')



"""
Some simple settings for use in other parts of the tool.
"""

import os

HOME = os.environ['HOME']
PYHOME_ROOT = os.path.join(HOME, '.config/pyhome')
PYHOME_REPO = os.path.join(PYHOME_ROOT, 'repos')



"""
Symbolic link creation and handling.
"""

from __future__ import print_function

import os
import shutil
from pyhome import settings

# Python 2/3 compatability
try:
    input = raw_input
except NameError:
    pass

# Files to search for that may contain subdir definitions
SUBDIR_FILENAMES = ['.subdir',
                    '.pyhome_subdir',
                    '.homesick_subdir']


def overwrite_prompt():
    userinput = input('Overwrite [yn]: ')
    clean = userinput.strip().lower()
    if clean == 'y':
        return True
    elif clean == 'n':
        return False
    else:
        print('Invalid input: "{}"'.format(input))
        return overwrite_prompt()


def create_single_symlink(linkpath, targetpath):
    """
    Create a symbolic link, printing status messages and prompting if needed.
    """

    create = True

    # Check current file status
    if os.path.exists(linkpath) or os.path.islink(linkpath):

        # Is it a link already
        if os.path.islink(linkpath):

            # Does it point to the right place?
            if (os.path.exists(linkpath)
                    and os.path.samefile(linkpath, targetpath)):
                print('Unchanged: {} -> {}'.format(linkpath, targetpath))
                return

            else:
                print('Conflict: {} -> {}'.format(linkpath, targetpath))
                print('          {}'.format(linkpath))
                print('              currently points to')
                print('          {}'.format(os.readlink(linkpath)))

                if overwrite_prompt():
                    os.unlink(linkpath)
                else:
                    create = False

        else:
            print('Conflict: {} -> {}'.format(linkpath, targetpath))
            print('          {}'.format(linkpath))

            if os.path.isdir(linkpath):
                print('              exists and is a directory')
                if overwrite_prompt():
                    shutil.rmtree(linkpath)
                else:
                    create = False

            else:
                print('              exists')
                if overwrite_prompt():
                    os.remove(linkpath)
                else:
                    create = False

    if create:
        os.symlink(targetpath, linkpath)
        print('Linked:   {} -> {}'.format(linkpath, targetpath))


def clear_single_symlink(linkpath, targetpath):
    """
    Remove a symbolic link, printing status messages and prompting if needed.
    """

    if not os.path.exists(linkpath):
        # Skip quietly
        pass

    elif not os.path.islink(linkpath):
        print('Skipping: {} -> {}'.format(linkpath, targetpath))
        print('          {}'.format(linkpath))
        print('              is not a link')

    elif not os.path.samefile(linkpath, targetpath):
        print('Skipping: {} -> {}'.format(linkpath, targetpath))
        print('          {}'.format(linkpath))
        print('              does not point to')
        print('          {}'.format(os.readlink(linkpath)))

    else:
        os.unlink(linkpath)
        print('Unlinked: {} -> {}'.format(linkpath, targetpath))


def splitall(path):
    """
    Split a path into its constituent parts.

    From the Python Cookbook: https://goo.gl/ooy0sx
    """

    allparts = []

    while 1:
        parts = os.path.split(path)

        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break

        elif parts[1] == path:  # sentinel for relative paths
            allparts.insert(0, parts[1])
            break

        else:
            path = parts[0]
            allparts.insert(0, parts[1])

    return allparts


def linkable_files(directory, subdirs=[]):
    """
    Generate a list of linkable files in the directory, accounting for subdir
    definitions.
    """

    # Get the first directory of each subdir
    subdir_top = [splitall(sd)[0] for sd in subdirs]

    for file in os.listdir(directory):

        if file in subdir_top:

            # Get any children
            children = []
            for top, sd in zip(subdir_top, subdirs):
                if file == top:
                    remaining = splitall(sd)[1:]
                    if len(remaining) > 0:
                        children.append(os.path.join(*remaining))

            for subfile in linkable_files(os.path.join(directory, file),
                                          children):
                yield os.path.join(file, subfile)

        else:
            yield file


def load_subdirs(repo):
    """
    Load any subdir definitions for this repo.
    """

    subdirs = []

    # Iterate over all possible subdir files
    for filename in SUBDIR_FILENAMES:
        fullpath = os.path.join(repo, filename)
        if os.path.exists(fullpath):

            # Load the subdirs
            with open(fullpath) as fp:
                for line in fp:
                    line = line.strip()
                    if len(line) > 0:
                        subdirs.append(line)

    return subdirs


def repo_symlink_map(repo, function):
    """
    Map a function to all linkable file (from, to) pairs in this repo.
    """

    home = os.path.join(repo, 'home')

    for fname in linkable_files(home, load_subdirs(repo)):
        linkpath = os.path.join(settings.HOME, fname)
        targetpath = os.path.join(home, fname)
        function(linkpath, targetpath)


def repo_create_symlinks(repo):
    """
    Generate symbolic links for this repo.
    """
    repo_symlink_map(repo, create_single_symlink)


def repo_clear_symlinks(repo):
    """
    Remove symbolic links for this repo.
    """
    repo_symlink_map(repo, clear_single_symlink)



from __future__ import print_function

import os
import sys
import shutil
import argparse
from glob import glob

from pyhome import settings, git, symlink


def repo_list():
    """
    Return list of all git repos in repodir
    """

    # Assuming all git repos contain a .git folder
    pattern = os.path.join(settings.PYHOME_REPO, '*', '.git')

    # Do some extra checks
    repos = []
    for path in glob(pattern):
        # .git should be a folder
        if not os.path.isdir(path):
            continue
        repo_root = os.path.dirname(path)
        repos.append(os.path.basename(repo_root))

    return sorted(repos)


def list(args):
    """
    List all existing repos in your pyhome
    """

    print('Current pyhome repos:')
    for repo in repo_list():
        print('    {}'.format(repo))


def init(args):
    """
    Initialise a new repo in your pyhome
    """

    # Make sure repo dir exists
    if not os.path.exists(settings.PYHOME_REPO):
        os.makedirs(settings.PYHOME_REPO)

    print('Initialising repo ...')
    git.init(settings.PYHOME_REPO, args.name)


def clone(args):
    """
    Clone a new repo to your pyhome
    """

    # Make sure repo dir exists
    if not os.path.exists(settings.PYHOME_REPO):
        os.makedirs(settings.PYHOME_REPO)

    print('Cloning repo from {} ...'.format(args.url))
    git.clone(settings.PYHOME_REPO, args.url, args.name, args.submodules)


def remove(args):
    """
    Remove a repo from your pyhome
    """

    repopath = os.path.join(settings.PYHOME_REPO, args.repo)

    if not os.path.exists(repopath):
        print('No repo "{}" found.'.format(args.repo))
        # This is an error
        sys.exit(1)

    if not args.force:
        # Make sure of no accidental deletions
        tpl = 'Really remove "{}"? Run again with "-f" to confirm.'
        print(tpl.format(repopath))

    else:
        # Actually delete the thing
        print('Removing repo {} ...'.format(repopath))
        print(' - Unlinking ...')
        symlink.repo_clear_symlinks(repopath)
        print(' - Removing directory ...')
        shutil.rmtree(repopath)


def pull(args):
    """
    Pull a repo and optionally update its submodules
    """

    for r in repos_from_arguments(args):
        print('\nPulling repo in {} ...'.format(r))
        git.pull(r, args.submodules)


def link(args):
    """
    Generate links for this repo in your $HOME folder
    """

    for r in repos_from_arguments(args):
        print('\nCreating symlinks for repo {} ...'.format(r))
        symlink.repo_create_symlinks(r)


def unlink(args):
    """
    Remove links for this repo in your $HOME folder
    """

    for r in repos_from_arguments(args):
        print('\nRemoving symlinks for repo {} ...'.format(r))
        symlink.repo_clear_symlinks(r)


def parser_add_repo_options(parser):
    """
    Add some standard options for selecting repos to a parser
    """

    parser.add_argument('repos', nargs='*', help='repos to link')
    parser.add_argument('-a', '--all', action='store_true',
                        help='link all repos')


def repos_from_arguments(args):
    """
    Build a list of repos rom the arguments defined in parser_add_repo_options
    """

    if args.all:
        repos = repo_list()

        if len(repos) == 0:
            print('No repos have yet been cloned to your pyhome')
            print('    Repo dir: {}'.format(settings.PYHOME_REPOS))
            sys.exit(1)

    else:
        repos = args.repos

        if len(repos) == 0:
            print('Either specify repos on the command line or use --all')
            sys.exit(1)

    return [os.path.join(settings.PYHOME_REPO, r) for r in repos]


def main():
    """
    Execute the main command line interface.
    """

    # Define the CLI
    description = 'A dotfile management and synchronisation tool.'
    parser = argparse.ArgumentParser(description=description)
    subparsers = parser.add_subparsers()

    # Fix for issue #9253
    # http://bugs.python.org/issue9253#msg186387
    subparsers.required = True
    subparsers.dest = 'subcommand'

    def add(func):
        sub = subparsers.add_parser(func.__name__, help=func.__doc__,
                                    description=func.__doc__)
        sub.set_defaults(func=func)
        return sub

    # List existing repos
    add(list)

    # Initialise a repo
    init_parser = add(init)
    init_parser.add_argument('name',
                             help='name of the new repo')

    # Clone a repo
    clone_parser = add(clone)
    clone_parser.add_argument('url',
                              help='URL of git repo to clone')
    clone_parser.add_argument('name',
                              nargs='?',
                              help='optionally specify name of cloned repo')
    clone_parser.add_argument('--no-submodules',
                              action='store_false', dest='submodules',
                              help='do not update submodules')

    # Remove a repo
    rm_parser = add(remove)
    rm_parser.add_argument('repo',
                           help='name of the repo to remove')
    rm_parser.add_argument('-f', '--force',
                           action='store_true',
                           help='confirm the removal of the repo')

    # Pull a repo/repos
    pull_parser = add(pull)
    parser_add_repo_options(pull_parser)
    pull_parser.add_argument('--no-submodules',
                             action='store_false', dest='submodules',
                             help='do not update submodules')

    # Set up symlinks
    link_parser = add(link)
    parser_add_repo_options(link_parser)

    # Clear symlinks
    unlink_parser = add(unlink)
    parser_add_repo_options(unlink_parser)

    args = parser.parse_args()
    args.func(args)


# Enable script execution with 'python -m pyhome'
if __name__ == '__main__':
    main()



from setuptools import setup, find_packages

setup(
    name='pyhome',
    version='0.3.0',
    description='Dotfile management and synchronisation tool',
    long_description=open('README.rst').read(),
    url='https://github.com/acroz/pyhome',
    author='Andrew Crozier',
    author_email='wacrozier@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: System Administrators',
        'Topic :: System :: Installation/Setup',
        'Topic :: System :: Systems Administration',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5'
    ],
    packages=find_packages(),
    entry_points={
        'console_scripts': [
          'pyhome=pyhome.__main__:main'
        ]
    }
)



"""
Directory context manager for the safe execution of a function with a temporary
changed working directory.
"""

import os
from contextlib import contextmanager


@contextmanager
def dircontext(directory):
    """
    Context manager for the execution of a code block inside another directory.
    """

    # Store the current directory
    orig_dir = os.getcwd()

    try:
        # Change the working directory
        os.chdir(directory)
        yield

    finally:
        # Resotre the original directory
        os.chdir(orig_dir)
